@@ -115,13 +115,12 @@ void libvisio::VSDSVGGenerator::setStyle(const ::WPXPropertyList &propList, cons
   m_gradient = gradient;
   if(m_style["draw:shadow"] && m_style["draw:shadow"]->getStr() == "visible")
   {
-    unsigned shadowColour = 0;
     double shadowRed = 0.0;
     double shadowGreen = 0.0;
     double shadowBlue = 0.0;
     if (m_style["draw:shadow-color"])
     {
-      shadowColour = stringToColour(m_style["draw:shadow-color"]->getStr());
+      unsigned shadowColour = stringToColour(m_style["draw:shadow-color"]->getStr());
       shadowRed = (double)((shadowColour & 0x00ff0000) >> 16)/255.0;
       shadowGreen = (double)((shadowColour & 0x0000ff00) >> 8)/255.0;
       shadowBlue = (double)(shadowColour & 0x000000ff)/255.0;
@@ -435,8 +434,34 @@ void libvisio::VSDSVGGenerator::drawGraphicObject(const ::WPXPropertyList &propL
   WPXString base64 = binaryData.getBase64Data();
   m_outputSink << "<svg:image ";
   if (propList["svg:x"] && propList["svg:y"] && propList["svg:width"] && propList["svg:height"])
-    m_outputSink << "x=\"" << doubleToString(72*(propList["svg:x"]->getDouble())) << "\" y=\"" << doubleToString(72*(propList["svg:y"]->getDouble())) << "\" ";
-  m_outputSink << "width=\"" << doubleToString(72*(propList["svg:width"]->getDouble())) << "\" height=\"" << doubleToString(72*(propList["svg:height"]->getDouble())) << "\" ";
+  {
+    double x(propList["svg:x"]->getDouble());
+    double y(propList["svg:y"]->getDouble());
+    double width(propList["svg:width"]->getDouble());
+    double height(propList["svg:height"]->getDouble());
+    bool flipX(propList["draw:mirror-horizontal"] && propList["draw:mirror-horizontal"]->getInt());
+    bool flipY(propList["draw:mirror-vertical"] && propList["draw:mirror-vertical"]->getInt());
+
+    double xmiddle = x + width / 2.0;
+    double ymiddle = y + height / 2.0;
+    m_outputSink << "x=\"" << doubleToString(72*x) << "\" y=\"" << doubleToString(72*y) << "\" ";
+    m_outputSink << "width=\"" << doubleToString(72*width) << "\" height=\"" << doubleToString(72*height) << "\" ";
+    m_outputSink << "transform=\"";
+    m_outputSink << " translate(" << doubleToString(72*xmiddle) << ", " << doubleToString (72*ymiddle) << ") ";
+    m_outputSink << " scale(" << (flipX ? "-1" : "1") << ", " << (flipY ? "-1" : "1") << ") ";
+    // rotation is around the center of the object's bounding box
+    if (propList["libwpg:rotate"])
+    {
+      double angle(propList["libwpg:rotate"]->getDouble());
+      while (angle > 180.0)
+        angle -= 360.0;
+      while (angle < -180.0)
+        angle += 360.0;
+      m_outputSink << " rotate(" << doubleToString(angle) << ") ";
+    }
+    m_outputSink << " translate(" << doubleToString(-72*xmiddle) << ", " << doubleToString (-72*ymiddle) << ") ";
+    m_outputSink << "\" ";
+  }
   m_outputSink << "xlink:href=\"data:" << propList["libwpg:mime-type"]->getStr().cstr() << ";base64,";
   m_outputSink << base64.cstr();
   m_outputSink << "\" />\n";
