@@ -57,8 +57,9 @@ libvisio::VSDXContentCollector::VSDXContentCollector(
   m_shadowOffsetX(0.0), m_shadowOffsetY(0.0),
   m_scale(1.0), m_x(0.0), m_y(0.0), m_originalX(0.0), m_originalY(0.0), m_xform(),
   m_txtxform(0), m_currentGeometry(), m_groupXForms(groupXFormsSequence[0]),
-  m_currentForeignData(), m_currentForeignProps(),
-  m_currentShapeId(0), m_foreignType(0), m_foreignFormat(0), m_styleProps(),
+  m_currentForeignData(), m_currentOLEData(), m_currentForeignProps(),
+  m_currentShapeId(0), m_foreignType(0), m_foreignFormat(0), m_foreignOffsetX(0.0),
+  m_foreignOffsetY(0.0), m_foreignWidth(0.0), m_foreignHeight(0.0), m_styleProps(),
   m_lineColour("black"), m_fillType("none"), m_linePattern(1),
   m_fillPattern(1), m_fillFGTransparency(0), m_fillBGTransparency(0),
   m_noLine(false), m_noFill(false), m_noShow(false), m_colours(), m_fonts(),
@@ -242,35 +243,8 @@ void libvisio::VSDXContentCollector::_fillAndShadowProperties(unsigned colourInd
   m_styleProps.insert("draw:fill", m_fillType);
 }
 
-void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour c, unsigned linePattern, unsigned startMarker, unsigned endMarker, unsigned lineCap)
+void libvisio::VSDXContentCollector::_applyLinePattern()
 {
-  m_linePattern = linePattern;
-
-  if (linePattern == 0) return; // No need to add style
-
-  m_styleProps.insert("svg:stroke-width", m_scale*strokeWidth);
-  m_lineColour = getColourString(c);
-  m_styleProps.insert("svg:stroke-color", m_lineColour);
-  if (c.a)
-    m_styleProps.insert("svg:stroke-opacity", (1 - c.a/255.0), WPX_PERCENT);
-  else
-    m_styleProps.insert("svg:stroke-opacity", 1.0, WPX_PERCENT);
-  switch (lineCap)
-  {
-  case 0:
-    m_styleProps.insert("svg:stroke-linecap", "round");
-    m_styleProps.insert("svg:stroke-linejoin", "round");
-    break;
-  case 2:
-    m_styleProps.insert("svg:stroke-linecap", "square");
-    m_styleProps.insert("svg:stroke-linejoin", "miter");
-    break;
-  default:
-    m_styleProps.insert("svg:stroke-linecap", "butt");
-    m_styleProps.insert("svg:stroke-linejoin", "miter");
-    break;
-  }
-
   int dots1 = 0;
   int dots2 = 0;
   double dots1len = 0.0;
@@ -278,7 +252,7 @@ void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour
   double gap = 0.0;
 
   m_styleProps.remove("draw:stroke");
-  switch (linePattern)
+  switch (m_linePattern)
   {
   case 2: // "6, 3"
     dots1 = dots2 = 1;
@@ -424,11 +398,11 @@ void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour
     break;
   }
 
-  if (linePattern == 0)
+  if (m_linePattern == 0)
     m_styleProps.insert("draw:stroke", "none");
-  else if (linePattern == 1)
+  else if (m_linePattern == 1)
     m_styleProps.insert("draw:stroke", "solid");
-  else if (linePattern > 1 && linePattern <= 23)
+  else if (m_linePattern > 1 && m_linePattern <= 23)
   {
     m_styleProps.insert("draw:stroke", "dash");
     m_styleProps.insert("draw:dots1", dots1);
@@ -441,6 +415,38 @@ void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour
     // FIXME: later it will require special treatment for custom line patterns
     // patt ID is 0xfe, link to stencil name is in 'Line' blocks
     m_styleProps.insert("draw:stroke", "solid");
+}
+
+void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour c, unsigned linePattern, unsigned startMarker, unsigned endMarker, unsigned lineCap)
+{
+  m_linePattern = linePattern;
+
+  if (linePattern == 0) return; // No need to add style
+
+  m_styleProps.insert("svg:stroke-width", m_scale*strokeWidth);
+  m_lineColour = getColourString(c);
+  m_styleProps.insert("svg:stroke-color", m_lineColour);
+  if (c.a)
+    m_styleProps.insert("svg:stroke-opacity", (1 - c.a/255.0), WPX_PERCENT);
+  else
+    m_styleProps.insert("svg:stroke-opacity", 1.0, WPX_PERCENT);
+  switch (lineCap)
+  {
+  case 0:
+    m_styleProps.insert("svg:stroke-linecap", "round");
+    m_styleProps.insert("svg:stroke-linejoin", "round");
+    break;
+  case 2:
+    m_styleProps.insert("svg:stroke-linecap", "square");
+    m_styleProps.insert("svg:stroke-linejoin", "miter");
+    break;
+  default:
+    m_styleProps.insert("svg:stroke-linecap", "butt");
+    m_styleProps.insert("svg:stroke-linejoin", "miter");
+    break;
+  }
+
+  _applyLinePattern();
 
   // Deal with line markers (arrows, etc.)
   if (startMarker > 0)
@@ -465,6 +471,7 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
   double x = 0;
   double y = 0;
   bool firstPoint = true;
+  bool wasMove = false;
 
   for (unsigned i = 0; i < m_currentGeometry.size(); i++)
   {
@@ -475,10 +482,11 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
       startX = x;
       startY = y;
       firstPoint = false;
+      wasMove = true;
     }
     else if (m_currentGeometry[i]["libwpg:path-action"]->getStr() == "M")
     {
-      if (((startX == x && startY == y) || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none")) && path.count())
+      if ((startX == x && startY == y) && path.count() && !wasMove)
       {
         WPXPropertyList closedPath;
         closedPath.insert("libwpg:path-action", "Z");
@@ -488,15 +496,17 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
       y = m_currentGeometry[i]["svg:y"]->getDouble();
       startX = x;
       startY = y;
+      wasMove = true;
     }
     else
     {
       x = m_currentGeometry[i]["svg:x"]->getDouble();
       y = m_currentGeometry[i]["svg:y"]->getDouble();
+      wasMove = false;
     }
     path.append(m_currentGeometry[i]);
   }
-  if (((startX == x && startY == y) || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none")) && path.count())
+  if ((startX == x && startY == y) && path.count() && !wasMove)
   {
     WPXPropertyList closedPath;
     closedPath.insert("libwpg:path-action", "Z");
@@ -578,7 +588,7 @@ void libvisio::VSDXContentCollector::_flushText()
   const unsigned char *pTextBuffer = m_textStream.getDataBuffer();
 
   for (std::vector<VSDXParaStyle>::iterator paraIt = m_paraFormats.begin();
-       paraIt < m_paraFormats.end() || charIndex < m_charFormats.size(); paraIt++)
+       paraIt < m_paraFormats.end() || charIndex < m_charFormats.size(); ++paraIt)
   {
     WPXPropertyList paraProps;
 
@@ -590,15 +600,21 @@ void libvisio::VSDXContentCollector::_flushText()
     switch ((*paraIt).align)
     {
     case 0: // left
-      paraProps.insert("fo:text-align", "left");
+      if (!(*paraIt).flags)
+        paraProps.insert("fo:text-align", "left");
+      else
+        paraProps.insert("fo:text-align", "end");
       break;
     case 2: // right
-      paraProps.insert("fo:text-align", "end");
+      if (!(*paraIt).flags)
+        paraProps.insert("fo:text-align", "end");
+      else
+        paraProps.insert("fo:text-align", "left");
       break;
-    case 4: // justify
+    case 3: // justify
       paraProps.insert("fo:text-align", "justify");
       break;
-    case 5: // full
+    case 4: // full
       paraProps.insert("fo:text-align", "full");
       break;
     default: // center
@@ -709,19 +725,111 @@ void libvisio::VSDXContentCollector::_flushText()
 
 void libvisio::VSDXContentCollector::_flushCurrentForeignData()
 {
-  m_currentForeignProps.insert("svg:width", m_scale*m_xform.width);
-  m_currentForeignProps.insert("svg:height", m_scale*m_xform.height);
-  double x = 0.0;
-  double y = 0.0;
-  transformPoint(x,y);
+  double x1 = m_foreignOffsetX;
+  double y1 = m_foreignOffsetY;
+  double x2 = m_foreignOffsetX + m_foreignWidth;
+  double y2 = m_foreignOffsetY + m_foreignHeight;
+  double x3 = x1;
+  double y3 = y2;
+  double x4 = x2;
+  double y4 = y1;
+  transformPoint(x1,y1);
+  transformPoint(x2,y2);
+  transformPoint(x3,y3);
+  transformPoint(x4,y4);
+  double angle = 0.0;
+  transformAngle(angle);
 
-  m_currentForeignProps.insert("svg:x", m_scale*x);
-  // Y axis starts at the bottom not top
-  m_currentForeignProps.insert("svg:y", m_scale*(y - m_xform.height));
+  double xmin = x1;
+  double xmax = x1;
+
+  if (xmin > x2)
+    xmin = x2;
+  if (xmax < x2)
+    xmax = x2;
+
+  if (xmin > x3)
+    xmin = x3;
+  if (xmax < x3)
+    xmax = x3;
+
+  if (xmin > x4)
+    xmin = x4;
+  if (xmax < x4)
+    xmax = x4;
+
+  double ymin = y1;
+  double ymax = y1;
+
+  if (ymin > y2)
+    ymin = y2;
+  if (ymax < y2)
+    ymax = y2;
+
+  if (ymin > y3)
+    ymin = y3;
+  if (ymax < y3)
+    ymax = y3;
+
+  if (ymin > y4)
+    ymin = y4;
+  if (ymax < y4)
+    ymax = y4;
+
+  double xmiddle = (xmax + xmin) / 2.0;
+  double ymiddle = (ymax + ymin) / 2.0;
+
+  bool flipX = false;
+  bool flipY = false;
+
+  // If any parent group is flipped, invert flips and
+  // revert the flip information that is already incorporated
+  // in the angle.
+  unsigned shapeId = m_currentShapeId;
+  while (true)
+  {
+    std::map<unsigned, XForm>::iterator iterX = m_groupXForms.find(shapeId);
+    if (iterX != m_groupXForms.end())
+    {
+      XForm xform = iterX->second;
+      if (xform.flipX)
+      {
+        flipX = !flipX;
+        angle = M_PI - angle;
+      }
+      if (xform.flipY)
+      {
+        flipY = !flipY;
+        angle *= -1.0;
+      }
+    }
+    else
+      break;
+    std::map<unsigned, unsigned>::iterator iter = m_groupMemberships.find(shapeId);
+    if (iter != m_groupMemberships.end())
+      shapeId = iter->second;
+    else
+      break;
+  }
+
+  WPXPropertyList styleProps(m_styleProps);
+
+  m_currentForeignProps.insert("svg:x", m_scale*(xmiddle - (m_foreignWidth / 2.0)));
+  m_currentForeignProps.insert("svg:width", m_scale*m_foreignWidth);
+  m_currentForeignProps.insert("svg:y", m_scale*(ymiddle - (m_foreignHeight / 2.0)));
+  m_currentForeignProps.insert("svg:height", m_scale*m_foreignHeight);
+
+  if (flipX)
+    m_currentForeignProps.insert("draw:mirror-horizontal", true);
+  if (flipY)
+    m_currentForeignProps.insert("draw:mirror-vertical", true);
+
+  if (angle != 0.0)
+    m_currentForeignProps.insert("libwpg:rotate", angle * 180 / M_PI, WPX_GENERIC);
 
   if (m_currentForeignData.size() && m_currentForeignProps["libwpg:mime-type"] && !m_noShow)
   {
-    m_shapeOutputDrawing->addStyle(m_styleProps, WPXPropertyListVector());
+    m_shapeOutputDrawing->addStyle(styleProps, WPXPropertyListVector());
     m_shapeOutputDrawing->addGraphicObject(m_currentForeignProps, m_currentForeignData);
   }
   m_currentForeignData.clear();
@@ -730,10 +838,10 @@ void libvisio::VSDXContentCollector::_flushCurrentForeignData()
 
 void libvisio::VSDXContentCollector::_flushCurrentPage()
 {
-  if (m_pageShapeOrder.size())
+  if (!m_pageShapeOrder.empty())
   {
     std::stack<std::pair<unsigned, VSDXOutputElementList> > groupTextStack;
-    for (std::list<unsigned>::iterator iterList = m_pageShapeOrder.begin(); iterList != m_pageShapeOrder.end(); iterList++)
+    for (std::list<unsigned>::iterator iterList = m_pageShapeOrder.begin(); iterList != m_pageShapeOrder.end(); ++iterList)
     {
       std::map<unsigned, unsigned>::iterator iterGroup = m_groupMemberships.find(*iterList);
       if (iterGroup == m_groupMemberships.end())
@@ -938,7 +1046,7 @@ void libvisio::VSDXContentCollector::collectInfiniteLine(unsigned /* id */, unsi
 
     xmove = points.begin()->first;
     ymove = points.begin()->second;
-    for (std::map<double, double>::iterator iter = points.begin(); iter != points.end(); iter++)
+    for (std::map<double, double>::iterator iter = points.begin(); iter != points.end(); ++iter)
     {
       if (iter->first != xmove || iter->second != ymove)
       {
@@ -984,6 +1092,20 @@ void libvisio::VSDXContentCollector::collectForeignData(unsigned /* id */, unsig
   _handleForeignData(binaryData);
 }
 
+void libvisio::VSDXContentCollector::collectOLEList(unsigned /* id */, unsigned level)
+{
+  _handleLevelChange(level);
+  m_currentForeignData.clear();
+  WPXBinaryData binaryData;
+  _handleForeignData(binaryData);
+}
+
+void libvisio::VSDXContentCollector::collectOLEData(unsigned /* id */, unsigned level, const WPXBinaryData &oleData)
+{
+  _handleLevelChange(level);
+  m_currentForeignData.append(oleData);
+}
+
 void libvisio::VSDXContentCollector::_handleForeignData(const WPXBinaryData &binaryData)
 {
   if (m_foreignType == 1 || m_foreignType == 4) // Image
@@ -1083,6 +1205,10 @@ void libvisio::VSDXContentCollector::_handleForeignData(const WPXBinaryData &bin
       }
     }
   }
+  else if (m_foreignType == 2)
+  {
+    m_currentForeignProps.insert("libwpg:mime-type", "object/ole");
+  }
 }
 
 void libvisio::VSDXContentCollector::collectGeometry(unsigned /* id */, unsigned level, unsigned char geomFlags)
@@ -1099,6 +1225,8 @@ void libvisio::VSDXContentCollector::collectGeometry(unsigned /* id */, unsigned
   if ((m_noFill != noFill) || (m_noLine != noLine) || (m_noShow != noShow) || m_isFirstGeometry)
     _flushCurrentPath();
 
+  _applyLinePattern();
+
   m_isFirstGeometry = false;
   m_noFill = noFill;
   m_noLine = noLine;
@@ -1224,7 +1352,7 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned /* id */, unsigned
 
   // Convert control points to static co-ordinates
   for (std::vector<std::pair<double, double> >::iterator it = controlPoints.begin();
-       it != controlPoints.end(); it++)
+       it != controlPoints.end(); ++it)
   {
     if (xType == 0) // Percentage
       (*it).first *= m_xform.width;
@@ -1513,11 +1641,15 @@ void libvisio::VSDXContentCollector::collectShapeId(unsigned /* id */, unsigned
   _handleLevelChange(level);
 }
 
-void libvisio::VSDXContentCollector::collectForeignDataType(unsigned /* id */, unsigned level, unsigned foreignType, unsigned foreignFormat)
+void libvisio::VSDXContentCollector::collectForeignDataType(unsigned /* id */, unsigned level, unsigned foreignType, unsigned foreignFormat, double offsetX, double offsetY, double width, double height)
 {
   _handleLevelChange(level);
   m_foreignType = foreignType;
   m_foreignFormat = foreignFormat;
+  m_foreignOffsetX = offsetX;
+  m_foreignOffsetY = offsetY;
+  m_foreignWidth = width;
+  m_foreignHeight = height;
 }
 
 void libvisio::VSDXContentCollector::collectPageProps(unsigned /* id */, unsigned level, double pageWidth, double pageHeight, double shadowOffsetX, double shadowOffsetY, double scale)
@@ -1546,6 +1678,10 @@ void libvisio::VSDXContentCollector::collectShape(unsigned id, unsigned level, u
 
   m_foreignType = 0; // Tracks current foreign data type
   m_foreignFormat = 0; // Tracks foreign data format
+  m_foreignOffsetX = 0.0;
+  m_foreignOffsetY = 0.0;
+  m_foreignWidth = 0.0;
+  m_foreignHeight = 0.0;
 
   m_originalX = 0.0;
   m_originalY = 0.0;
@@ -1611,6 +1747,10 @@ void libvisio::VSDXContentCollector::collectShape(unsigned id, unsigned level, u
       {
         m_foreignType = m_stencilShape->m_foreign->type;
         m_foreignFormat = m_stencilShape->m_foreign->format;
+        m_foreignOffsetX = m_stencilShape->m_foreign->offsetX;
+        m_foreignOffsetY = m_stencilShape->m_foreign->offsetY;
+        m_foreignWidth = m_stencilShape->m_foreign->width;
+        m_foreignHeight = m_stencilShape->m_foreign->height;
         _handleForeignData(m_stencilShape->m_foreign->data);
       }
 
@@ -1749,10 +1889,10 @@ void libvisio::VSDXContentCollector::collectText(unsigned /*id*/, unsigned level
 }
 
 void libvisio::VSDXContentCollector::collectVSDXParaStyle(unsigned /* id */ , unsigned level, unsigned charCount, double indFirst, double indLeft, double indRight,
-    double spLine, double spBefore, double spAfter, unsigned char align)
+    double spLine, double spBefore, double spAfter, unsigned char align, unsigned flags)
 {
   _handleLevelChange(level);
-  VSDXParaStyle format(charCount, indFirst, indLeft, indRight, spLine, spBefore, spAfter, align);
+  VSDXParaStyle format(charCount, indFirst, indLeft, indRight, spLine, spBefore, spAfter, align, flags);
   m_paraFormats.push_back(format);
 }
 
@@ -1840,7 +1980,7 @@ void libvisio::VSDXContentCollector::collectCharIXStyle(unsigned /*id*/ , unsign
 }
 
 void libvisio::VSDXContentCollector::collectParaIXStyle(unsigned /* id */, unsigned level, unsigned /* charCount */, double /* indFirst */, double /* indLeft */, double /* indRight */,
-    double /* spLine */, double /* spBefore */, double /* spAfter */, unsigned char /* align */)
+    double /* spLine */, double /* spBefore */, double /* spAfter */, unsigned char /* align */, unsigned /* flags */)
 {
   _handleLevelChange(level);
 }
@@ -1949,7 +2089,7 @@ void libvisio::VSDXContentCollector::_handleLevelChange(unsigned level)
         m_NURBSData = m_stencilShape->m_nurbsData;
         m_polylineData = m_stencilShape->m_polylineData;
 
-        if (m_currentGeometry.size() == 0)
+        if (m_currentGeometry.empty())
         {
           for (unsigned i = 0; i < m_stencilShape->m_geometries.size(); i++)
           {
